# Copyright 2016 The Johns Hopkins University Applied Physics Laboratory
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import boto3
import os
import json
import time

class QueueRecovery(object):
    """Class to manage recovering data from a queue"""

    def __init__(self, queue_name, region="us-east-1"):
        self.sqs = boto3.resource('sqs', region_name=region)
        self.queue = self.sqs.Queue(url=queue_name)

    def simple_store_messages(self, output_dir):
        """Method to store all remaining messages in a queue for later use during a recovery/debug operation

        Currently this assumes you can download all messages BEFORE the visibility timeout. Otherwise you will enter an
        endless loop. For large number of messages, additional development will be needed.

        Args:
            output_dir(str): directory to dump data

        Returns:
            None
        """
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        cnt = 0
        while True:
            msgs = self.queue.receive_messages(MaxNumberOfMessages=10, WaitTimeSeconds=10)

            if msgs:
                for msg in msgs:
                    cnt += 1
                    with open(os.path.join(output_dir, "{}.json".format(msg.message_id)), "wt") as msg_file:
                        msg_file.write(msg.body)
            else:
                break

        print("Saved {} messages to {}.".format(cnt, output_dir))

    def restore_messages(self, input_dir):
        """Method to re-load a backed up messages to an ingest queue"""
        for msg_file in os.listdir(input_dir):
            print(msg_file)
            with open(os.path.join(input_dir, msg_file), "rt") as msg:
                msg_body = msg.read()
                response = self.queue.send_message(MessageBody=msg_body)
                if response["ResponseMetadata"]["HTTPStatusCode"] != 200:
                    print("failed to upload {}".format(msg_file))

    def invoke_ingest(self, input_dir, x_tile, y_tile):
        """Method to trigger lambda functions until ingest completes"""
        # Load a single message to build the object metadata
        filename = [x for x in os.listdir(input_dir)][0]
        with open(os.path.join(input_dir, filename), "rt") as msg:
            metadata = json.load(msg)

        metadata["tile_size_x"] = x_tile
        metadata["tile_size_y"] = y_tile
        metadata["lambda-name"] = "ingest"

        # Get how many to invoke
        starting_message_count = int(self.queue.attributes['ApproximateNumberOfMessages'])
        num_invocations = range(0, starting_message_count)
        print("Triggering {} lambdas".format(starting_message_count))

        # Invoke Ingest lambda functions
        lambda_client = boto3.client('lambda', region_name="us-east-1")
        cnt = 0
        throttle_count = 30
        for _ in num_invocations:
            lambda_client.invoke(FunctionName=metadata["parameters"]["ingest_lambda"],
                                 InvocationType='Event',
                                 Payload=json.dumps(metadata).encode())
            cnt += 1
            if cnt >= throttle_count:
                print("Invoked {}...throttling...".format(throttle_count))
                time.sleep(5)
                cnt = 0

        print("Waiting for 2.5 minutes message timeout to check outcome...")
        time.sleep(150)
        print("Started with {} messages. Resulted in {} messages".format(starting_message_count,
                                                                         self.queue.attributes['ApproximateNumberOfMessages']))



# Allen Institute's plugin to upload coregisterd data
# This plugin should work with the config files in configs/coregistered/ folder
from __future__ import absolute_import
import os
import logging
import boto3
from botocore.exceptions import ClientError

from .path import PathProcessor
from .tile import TileProcessor
from io import BytesIO
import six
import numpy as np
from PIL import Image


class AiPathProcessor(PathProcessor):
    def setup(self, parameters):
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):

        filename = "{}/{}/{}.{}".format(z_index, y_index, x_index, self.parameters["filetype"])
        full_path = os.path.join(self.parameters["root_dir"], filename)
        return full_path
        # if os.path.isfile(full_path):
        #     return full_path
        # else:
        #     return ''


class AiS3TileProcessor(TileProcessor):
    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.data = None
        self.logger = logging.getLogger('ingest-client')

    def setup(self, parameters):
        self.parameters = parameters
        self.client = boto3.client("s3")

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        try:
            resp = self.client.get_object(Bucket=self.parameters['bucket'],
                                          Key=file_path)
            if resp['ResponseMetadata']['HTTPStatusCode'] != 200:
                self.logger.error('HTTPStatusCode: {} on {}'.format(resp['ResponseMetadata']['HTTPStatusCode'],
                                                                    file_path))
                return None
            else:
                #return resp['Body']
                #data = np.load(BytesIO(resp['Body'].read()))
                # Save sub-img to png and return handle
                # upload_img = Image.fromarray(np.squeeze(data))
                # output = six.BytesIO()
                #upload_img.save(output, format="PNG")

                data = BytesIO(resp['Body'].read())
                # Send handle back
                return data
        except ClientError as ex:
            if ex.response['Error']['Code'] == 'NoSuchKey':
                self.logger.warning('No object found: {}'.format(file_path))
                return open('/ingest-client/black-1024x1024.png', 'rb')
            elif ex.response['Error']['Code'] == 'NoSuchBucket':
                print('No such bucket found')
                self.logger.error('No bucket found: {}'.format(self.parameters['bucket']))
                raise ex
            else:
                raise ex

        # # if an image is missing, just send a blank one.
        # if os.path.isfile(file_path):
        #     logger = logging.getLogger('ingest-client')
        #     logger.info("Sending: {}".format(file_path))
        #     return open(file_path, 'rb')
        # else:
        #     logger = logging.getLogger('ingest-client')
        #     # logger.warning("Could not find path: {}".format(file_path))
        #     # TODO: figure out if I can return null here in order to avoid sending a blank png. How will an empty tile render using the API?
        #     return open('/allen/programs/celltypes/workgroups/em-connectomics/russelt/renderings/EM_Phase1/trakem2_PEA_fullstack_Trminusplus/1024x1024/empty.png', 'rb')


# Allen Institute's plugin to upload coregisterd data
# This plugin should work with the config files in configs/coregistered/ folder
from __future__ import absolute_import
import os
import logging

from .path import PathProcessor
from .tile import TileProcessor


class AiPathProcessor(PathProcessor):
    def setup(self, parameters):
        self.parameters = parameters

    def process(self, x_index, y_index, z_index, t_index=None):

        filename = "{}/{}/{}.{}".format(z_index, y_index, x_index, self.parameters["filetype"])
        full_path = os.path.join(self.parameters["root_dir"], filename)
        if os.path.isfile(full_path):
            return full_path
        else:
            return ''


class AiTileProcessor(TileProcessor):
    def __init__(self):
        """Constructor to add custom class var"""
        TileProcessor.__init__(self)
        self.data = None

    def setup(self, parameters):
        self.parameters = parameters

    def process(self, file_path, x_index, y_index, z_index, t_index=0):
        # if an image is missing, just send a blank one.
        if os.path.isfile(file_path):
            logger = logging.getLogger('ingest-client')
            logger.info("Sending: {}".format(file_path))
            return open(file_path, 'rb')
        else:
            logger = logging.getLogger('ingest-client')
            # logger.warning("Could not find path: {}".format(file_path))
            # TODO: figure out if I can return null here in order to avoid sending a blank png. How will an empty tile render using the API?
            return open('/allen/programs/celltypes/workgroups/em-connectomics/russelt/renderings/EM_Phase1/trakem2_PEA_fullstack_Trminusplus/1024x1024/empty.png', 'rb')


__version__ = '0.9.9'


def check_version():
    """
    Tells you if you have an old version of boss-ingest.
    """
    import requests
    r = requests.get('https://pypi.python.org/pypi/boss-ingest/json').json()
    r = r['info']['version']
    if r != __version__:
        print("You are using version {}. A newer version of boss-ingest is available: {} ".format(__version__, r) +
              "\n\n'pip install -U boss-ingest' to update.")
    else:
        print("You are using version {}, which is the latest available.".format(__version__))




import sys
import json
from .log import always_log_info
import pprint


def print_estimated_job(config_file=None, configuration=None):
    """Method to print details about the job the user is about to start

    Kwd Args:
        config_file(str): Path to the config file
        configuration(Configuration): Configuration object

    Returns:
        None
    """
    if config_file is not None:
        # Load file
        with open(config_file, 'rt') as cf:
            config = json.load(cf)
    elif configuration is not None:
        config = configuration.config_data
    else:
        raise Exception('Must provide either a configuration object or an absolute path to a config file')

    # Compute number of tiles
    num_x_tiles = config["ingest_job"]["extent"]["x"][1] - config["ingest_job"]["extent"]["x"][0]
    num_y_tiles = config["ingest_job"]["extent"]["y"][1] - config["ingest_job"]["extent"]["y"][0]
    num_z_tiles = config["ingest_job"]["extent"]["z"][1] - config["ingest_job"]["extent"]["z"][0]
    num_t_tiles = config["ingest_job"]["extent"]["t"][1] - config["ingest_job"]["extent"]["t"][0]

    num_tiles = (num_x_tiles * num_y_tiles) / (config["ingest_job"]["tile_size"]["x"] * config["ingest_job"]["tile_size"]["y"])
    num_tiles = num_tiles * num_z_tiles * num_t_tiles

    # Build Message
    pp = pprint.PrettyPrinter(indent=2)
    msg = "\n\n#### INGEST JOB SUMMARY ####\n\n"
    msg += "Data will be loaded into the Boss here:\n"
    msg += "  Collection: {}\n".format(config["database"]["collection"])
    msg += "  Experiment: {}\n".format(config["database"]["experiment"])
    msg += "  Channel: {}\n\n".format(config["database"]["channel"])
    msg += "Path Processor Configuration:\n"
    msg += "  Plugin: {}\n".format(config["client"]["path_processor"]["class"])
    msg += "  Parameters: {}\n".format(pp.pformat(config["client"]["path_processor"]["params"]).replace("\n", "\n              "))
    msg += "\nTile Processor Configuration:\n"
    msg += "  Plugin: {}\n".format(config["client"]["tile_processor"]["class"])
    msg += "  Parameters: {}\n".format(pp.pformat(config["client"]["tile_processor"]["params"]).replace("\n", "\n              "))
    msg += "\nTotal Number of Image Tiles to Upload: {}".format(int(num_tiles))

    # Print/Log
    always_log_info(msg)


class WaitPrinter(object):
    """Simple class to handle a print while waiting
    """
    def __init__(self):
        self.first_print = True
        self.wait_char = "."

    def print_msg(self, msg):
        """Method to print an initial message"""
        if self.first_print:
            sys.stdout.write("{}{}{}{}".format(msg, self.wait_char, self.wait_char, self.wait_char))
            sys.stdout.flush()
            self.first_print = False
        else:
            sys.stdout.write(self.wait_char)
            sys.stdout.flush()

    def finished(self, msg=None):
        """Method to print final message"""
        if msg:
            print(msg)
        else:
            print("Complete")



from setuptools import setup, find_packages
from codecs import open
from os import path
from ingestclient import __version__

# to update
# python setup.py sdist bdist_wheel
# twine upload --skip-existing dist/*


here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# get the dependencies and installs
with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '') for x in all_reqs if x.startswith('git+')]

setup(
    name='boss-ingest',
    version=__version__,

    description='Ingest client for the Boss',
    long_description=long_description,

    install_requires=install_requires,
    dependency_links=dependency_links,
    
    author='Johns Hopkins University Applied Physics Laboratory',
    author_email='iarpamicrons@jhuapl.edu',

    entry_points={
        'console_scripts': ['boss-ingest=ingestclient.client:main'],
    },
    #packages=find_packages('ingestclient'),
    packages=['ingestclient',
              'ingestclient.core',
              'ingestclient.plugins',
              'ingestclient.utils',
              'ingestclient.configs',
              'ingestclient.schema'],
    package_data={
        '': ['*.json'],
    },
    include_package_data=True,

    url='https://github.com/jhuapl-boss/ingest-client',
    #download_url='https://github.com/jhuapl-boss/ingest-client/tarball/' + __version__,

    license='Apache 2.0',
    classifiers=[
      'Development Status :: 4 - Beta',
      'Programming Language :: Python :: 3.5',
      'Programming Language :: Python :: 3.4',
      'Programming Language :: Python :: 2.7',
    ],
    keywords=[
        'brain',
        'microscopy',
        'neuroscience',
        'connectome',
        'connectomics',
        'spatial',
        'EM',
        'electron',
        'calcium',
        'database',
        'boss',
        'microns',
        'iarpa',
        'jhu'
    ]
)
